"""Numba-accelerated implementation of Donald Knuth's dancing cells algorithm."""

import numpy as np
from numba import njit


@njit(cache=True)
def algorithm_c(options, options_ptr, colors, n_items, n_secondary_items):
    """
    Donald Knuth's dancing cells algorithm C.
    Exact covering with colors.

    Parameters
    ----------
    options: array of integers, giving all options
    options_ptr: array of integers, pointing where each option
                 begins and ends
    colors: the color of each item in each option
    n_items: the number of possible items
    n_secondary_items: the number of items that are secondary

    Returns
    -------
    a generator object that yields solutions to the exact cover problem
    """

    # C1: Initialisation of the matrix
    n_data = len(options)
    n_opts = len(options_ptr) - 1
    n_primary_items = n_items - n_secondary_items

    # options_j gives the option index of each node in the matrix
    options_j = np.empty(n_data, dtype=np.int64)
    for j, i in enumerate(range(n_opts)):
        options_j[options_ptr[i] : options_ptr[i + 1]] = j

    # matrix_size gives the number of (active) options for each item
    matrix_size = np.zeros(n_items, dtype=np.int64)
    for node in range(n_data):
        matrix_size[options[node]] += 1

    # matrix_start_ptr points to the start of the "column" in the matrix
    # for each item
    matrix_start_ptr = np.empty(n_items, dtype=np.int64)
    matrix_start_ptr[0] = 0
    matrix_start_ptr[1:] = np.cumsum(matrix_size)[:-1]

    # matrix_set and matrix_loc are sparse-set partners
    matrix_set = np.empty(n_data, dtype=np.int64)
    matrix_loc = np.empty(n_data, dtype=np.int64)
    counts = np.zeros(n_items, dtype=np.int64)
    for node in range(n_data):
        i = options[node]
        val = matrix_start_ptr[i] + counts[i]
        matrix_loc[node] = val
        matrix_set[val] = node
        counts[i] += 1

    # the active items (items left to cover) use another sparse set
    # matrix_active_items and matrix_active_items_sparse are partners
    matrix_active_items = np.arange(n_items, dtype=np.int64)
    matrix_active_items_sparse = np.arange(n_items, dtype=np.int64)
    matrix_active_items_len = np.empty(1, dtype=np.int64)
    matrix_active_items_len[0] = n_items
    matrix_old_active_items_len = np.empty(1, dtype=np.int64)
    matrix_active_items_len[0] = n_items

    def active_insert(item, index):
        """Insert an item into position index in the active_items sparse set"""
        matrix_active_items[index] = item
        matrix_active_items_sparse[item] = index

    def deactivate_item(item):
        """C3: make an item inactive: remove from active_items list"""
        end_index = matrix_active_items_len[0] - 1
        end_item = matrix_active_items[end_index]
        index = matrix_active_items_sparse[item]
        active_insert(end_item, index)
        active_insert(item, end_index)
        matrix_active_items_len[0] -= 1

    def active_options(item):
        """the active options of a given item"""
        return matrix_set[
            matrix_start_ptr[item] : (matrix_start_ptr[item] + matrix_size[item])
        ]

    def remove_node(node):
        """remove a node from the matrix"""
        item = options[node]
        loc = matrix_loc[node]

        end_loc = matrix_start_ptr[item] + matrix_size[item] - 1
        end_node = matrix_set[end_loc]

        matrix_set[loc] = end_node
        matrix_set[end_loc] = node
        matrix_loc[end_node] = loc
        matrix_loc[node] = end_loc
        matrix_size[item] -= 1

    def hide(item, col, initial):
        """given an item and a coloring col, remove the relevant nodes"""
        # initial refers to whether we're hiding directly after a choose operation
        for node in active_options(item):
            if col == 0 or colors[node] != col:
                j = options_j[node]
                for k in range(options_ptr[j], options_ptr[j + 1]):
                    iprime = options[k]
                    if (
                        iprime != item
                        and matrix_active_items_sparse[iprime]
                        < matrix_old_active_items_len[0]
                    ):
                        if (
                            not initial
                            and matrix_size[iprime] == 1
                            and matrix_active_items_sparse[iprime]
                            < matrix_active_items_len[0]
                            and iprime < n_primary_items
                        ):
                            return False  # end if about to delete last
                        remove_node(k)
        return True

    def cover(node, item):
        """C6 and C7: main cover routine"""
        option = options_j[node]
        ptr_range = range(options_ptr[option], options_ptr[option + 1])
        matrix_old_active_items_len[0] = matrix_active_items_len[0]

        # C6: deactivate other items of option
        for ptr in ptr_range:
            itm = options[ptr]
            if (
                itm != item
                and matrix_active_items_sparse[itm] < matrix_active_items_len[0]
            ):
                deactivate_item(itm)

        # C7: hiding nodes
        for ptr in ptr_range:
            itm = options[ptr]
            col_hide = colors[ptr]
            if itm != item:
                if (
                    itm < n_primary_items
                    or matrix_active_items_sparse[itm] < matrix_old_active_items_len[0]
                ):
                    status = hide(itm, col_hide, False)
                    if status is False:
                        return -1
        return option

    def save_state():
        """C5: Save the current state (sizes) for backtracking"""
        return (matrix_size[:].copy(), matrix_active_items_len[0])

    def undo(state):
        """Restore a previous state of the matrix"""
        matrix_size[:], matrix_active_items_len[0] = state

    def choose():
        """C2: Choose the next item to cover. Return -1 if solved already"""
        # Using the minimum remaining value (MRV) heuristic here
        active_items = matrix_active_items[0 : matrix_active_items_len[0]]
        lens = matrix_size[active_items]
        cond = active_items >= n_primary_items
        lens[cond] = n_data
        if sum(lens[~cond]) == 0:
            return -1  # matrix is solved
        return matrix_active_items[np.argmin(lens)]

    # Main loop. A depth-first search, written here using a stack
    # rather than recursive as numba doesn't support yield from
    solution = []  # current solution
    node_stack = [[-1]]  # current list of nodes to explore (-1 is root)
    item_stack = [-1]  # current list of covered items
    initial_state = save_state()  # saved state for backtracking
    state_stack = [initial_state, initial_state]  # stack of states

    while node_stack:
        if len(node_stack[-1]) == 0:
            # backtracking, C10, C11
            node_stack.pop()
            state_stack.pop()
            item_stack.pop()
            undo(state_stack[-1])  # return to previous state, C11
            if solution:
                solution.pop()
        else:
            node = node_stack[-1].pop()  # C6
            if node >= 0:
                option = cover(node, item_stack[-1])  # C6 and C7
            else:
                option = -2
            if option == -1:  # case where cover failed
                undo(state_stack[-1])  # return to previous state, C11
            else:
                if option >= 0:
                    solution.append(option)  # include option in partial solution
                item = choose()  # C2
                if item == -1:
                    yield list(solution)  # found a solution!
                    solution.pop()
                    undo(state_stack[-1])  # return to previous state, C11
                else:
                    item_stack.append(item)
                    deactivate_item(item)  # C3
                    matrix_old_active_items_len[0] = matrix_active_items_len[0]
                    hide(item, 0, True)  # C4
                    state_stack.append(save_state())  # C5
                    node_stack.append(list(active_options(item)))



"""
xcover 
======

An exact cover with colors package.

The main function provided by this packages is "covers" which returns
a generator yielding solutions to exact cover problems.

"""

__version__ = "0.1.0"
__author__ = "John F. Rudge"

from .solvers import covers



"""Some useful additional routines"""
import numpy as np
from .solvers import items


def verify_exact_cover(solution, options, primary=None, secondary=None, colored=False):
    """
    Throws an exception if a solution does not represent an exact cover.
    """
    # A single list of all the options
    all_opts = [o for opt in options for o in opt]

    # Work out explicit primary and secondary item lists
    primary, secondary = items(all_opts, primary, secondary, colored)

    # Check the primary items for exact cover
    primary_options = [o for s in solution for o in options[s] if o in primary]
    cover = set().union(*[options[s] for s in solution])
    primary_cover = set(primary).intersection(cover)

    is_exact = len(primary_options) == len(primary_cover)
    is_cover = primary_cover == set(primary)
    assert is_exact
    assert is_cover

    # Check secondary items for covering at most once
    if colored:
        secondary_options = [
            o for s in solution for o in options[s] if o.split(":")[0] in secondary
        ]
    else:
        secondary_options = [o for s in solution for o in options[s] if o in secondary]

    # Count occurances of each secondary item
    counts = {s: 0 for s in secondary}
    colors = {s: 0 for s in secondary}
    for item in secondary_options:
        if colored:
            itm = item.split(":")[0]
            if len(item.split(":")) > 1:
                col = item.split(":")[1]
            else:
                col = 0
        else:
            itm, col = item, 0
        if col != 0:
            if colors[itm] == 0:
                colors[itm] = col
                counts[itm] += 1
            elif colors[itm] != col:
                assert False  # can't color twice
            else:
                continue  # fine to have another item of same color
        else:
            counts[itm] += 1
    for _, count in counts.items():
        assert count <= 1  # each only covered once


def bool_array_to_options(bool_array):
    """Convert a boolean array into a options list of lists"""
    return [list(np.nonzero(bool_array[i, :])[0]) for i in range(bool_array.shape[0])]



"""Simple file i/o"""


def read_xcover_from_file(filename):
    """
    Simple parser for files decribing exact cover problems in Knuth's format
    """
    with open(filename, "r", encoding="utf-8") as cover_file:
        first = True
        options = []
        for line in cover_file.readlines():
            if first:
                if (
                    line[0] == "|" or line[0] == "/" or line == "\n"
                ):  # a comment line, ignore
                    continue
                first = False
                split_pri_sec = line.rstrip("\n").split("|")
                primary = split_pri_sec[0].split(" ")
                if "" in primary:
                    primary.remove("")
                if len(split_pri_sec) > 1:
                    secondary = split_pri_sec[1].split(" ")
                    secondary.remove("")
                    colored = True
                else:
                    secondary = None
                    colored = False
            else:
                splits = line.rstrip("\n").split(" ")
                options.append(splits)

    return options, primary, secondary, colored



"""Solvers for exact cover problems"""

import numpy as np
from .dancing_cells import algorithm_c


def covers(options, primary=None, secondary=None, colored=False):
    """
    Exact cover with colors solver

    Parameters
    ----------
    options: a list of lists of items
    primary: a list of items that are primary (must be covered)
             if None, infer from the given options and secondary
    secondary: a list of items that are secondary (may be covered)
               if None, infer from the given options and primary
    colored: whether to do a colored solve (by default do not)
             for a colored solve, secondary items must be strings
             with colors in options separated by colons
             e.g. 'q:RED'

    Returns
    -------
    A generator object that yields solutions to the exact cover problem.
    Each yielded result is a list of integers specifying the indices of
    the chosen options.
    """
    options, options_ptr, colors, n_items, n_secondary = input_as_arrays(
        options,
        primary=primary,
        secondary=secondary,
        colored=colored,
    )
    yield from algorithm_c(options, options_ptr, colors, n_items, n_secondary)


def items(all_opts, primary=None, secondary=None, colored=False):
    """
    Infer the specific primary and secondary items and colors given user input.
    """
    if primary is None and secondary is None:
        # if items not specified assume all items in options are primary
        primary = set().union(all_opts)
        secondary = []
    elif primary is None:
        # if only secondary items specified, assume all others in options primary
        options_universe = set().union(all_opts)
        primary = options_universe.difference(secondary)
    elif secondary is None:
        # if only primary items specified, assume all others in options are secondary
        options_universe = set().union(all_opts)
        secondary = options_universe.difference(primary)
        if colored:
            secondary = {s.split(":")[0] for s in secondary}

    return primary, secondary


def color_indices(all_opts):
    """
    Label the colors of options by integer values.
    """
    col_names = {x.split(":")[-1] for x in all_opts if ":" in x}
    col_to_idx = {col: i + 1 for i, col in enumerate(col_names)}
    return np.array([col_to_idx[x.split(":")[-1]] if ":" in x else 0 for x in all_opts])


def input_as_arrays(options, primary=None, secondary=None, colored=False):
    """
    Convert the user-supplied input into array form for use in main algorithm
    """

    # A single list of all the options
    all_opts = [o for opt in options for o in opt]

    # Work out explicit primary and secondary item lists
    primary, secondary = items(all_opts, primary, secondary, colored)
    n_primary = len(primary)
    n_secondary = len(secondary)
    n_items = n_primary + n_secondary

    # Work out the color indices of each node
    if colored:
        colors = color_indices(all_opts)
    else:
        colors = np.zeros(len(all_opts), dtype=np.int64)

    # Form pointer array which gives start and end index of each option
    lens = np.array([len(opt) for opt in options])
    options_ptr = np.zeros(len(lens) + 1, dtype=np.int64)
    options_ptr[1:] = np.cumsum(lens)

    # dictionaries which enumerate each of the items
    primary_item_to_idx = {item: i for i, item in enumerate(primary)}
    secondary_item_to_idx = {item: i + n_primary for i, item in enumerate(secondary)}
    item_to_idx = {**primary_item_to_idx, **secondary_item_to_idx}

    # form final options as an array using the enumerated options
    if colored:
        enum_opts = [item_to_idx[x.split(":")[0]] for x in all_opts]
    else:
        enum_opts = [item_to_idx[x] for x in all_opts]
    options_as_array = np.hstack(enum_opts)

    return options_as_array, options_ptr, colors, n_items, n_secondary
